import unittest
import pygame

""" Collider registers Collidables to test collisions.

    A Collidable has a get_bounding_box() method that returns a bounding box.
"""

TRIGGER = 1
SOLID = 2


class Collider:
    def __init__(self):
        self._collidables = []
        self._update_rect_list()

    def _update_rect_list(self):
        self._rect_list = []
        for c in self._collidables:
            self._rect_list.append(c.get_bounding_box())
        self._needs_update = False

    def register_collidable(self, collidable):
        self._collidables.append(collidable)
        self._needs_update = True

    def update_collidable(self, collidable):
        if collidable in self._collidables:
            self._needs_update = True

    def test_collision(self, rect_to_test):
        if self._needs_update:
            self._update_rect_list()

        index_list = rect_to_test.collidelistall(self._rect_list)

        found_solid = False
        for index in index_list:
            collided = self._collidables[index]

            solid = collided.get_collide_flags() & SOLID == SOLID
            found_solid |= solid

            trigger = collided.get_collide_flags() & TRIGGER == TRIGGER
            if trigger:
                collided.on_triggered()

        return found_solid


class StaticBoundingBlock:
    def __init__(self, rect, trigger = None):
        self._rect = rect
        self._trigger = trigger
        self._flags = SOLID

        if trigger is not None:
            self._flags |= TRIGGER

    def get_bounding_box(self):
        return self._rect

    def get_collide_flags(self):
        return self._flags

    def on_triggered(self):
        if self._trigger is not None:
            self._trigger()


class MockCollidable():
    def __init__(self):
        self.bbox = pygame.Rect(5, 5, 4, 4)
        self.flag = SOLID 
        self.triggered = False

    def get_bounding_box(self):
        return self.bbox

    def get_collide_flags(self):
        return self.flag

    def on_triggered(self):
        self.triggered = True


class TestCollider(unittest.TestCase):
    def setUp(self):
        self.collidable = MockCollidable()
        self.collider = Collider()
        self.collider.register_collidable(self.collidable)

    def test_collision(self):
        is_colliding = self.collider.test_collision(pygame.Rect(0, 0, 10, 10))
        self.assertTrue(is_colliding)

    def test_no_collision(self):
        is_colliding = self.collider.test_collision(pygame.Rect(10, 10, 10, 10))
        self.assertFalse(is_colliding)

    def test_collision_after_movement(self):
        self.collidable.bbox = pygame.Rect(15, 15, 4, 4)
        self.collider.update_collidable(self.collidable)

        is_colliding = self.collider.test_collision(pygame.Rect(0, 0, 10, 10))
        self.assertFalse(is_colliding)

        is_colliding = self.collider.test_collision(pygame.Rect(10, 10, 10, 10))
        self.assertTrue(is_colliding)

    def test_trigger(self):
        self.collidable.flag = TRIGGER

        self.assertFalse(self.collidable.triggered)

        is_colliding = self.collider.test_collision(pygame.Rect(0, 0, 10, 10))
        self.assertFalse(is_colliding)

        self.assertTrue(self.collidable.triggered)

    def test_solid_trigger(self):
        self.collidable.flag = TRIGGER | SOLID

        self.assertFalse(self.collidable.triggered)

        is_colliding = self.collider.test_collision(pygame.Rect(0, 0, 10, 10))
        self.assertTrue(is_colliding)

        self.assertTrue(self.collidable.triggered)

if __name__ == '__main__':
    unittest.main()
