import pygame
import level
from .mover import Mover
from .character import Character, get_character_renderer
from .eventdispatcher import EventDispatcher
from .collider import Collider
from .textwriter import TextWriter
from .action import Action
from .collider import StaticBoundingBlock

game_surface_size = width, height = 800, 600
WHITE = pygame.Color(255, 255, 255)
BLACK = pygame.Color(0, 0, 0)
BLUE = pygame.Color(0, 0, 128)


class LeftBorder(StaticBoundingBlock):
    def __init__(self):
        StaticBoundingBlock.__init__(self, pygame.Rect(-32, 0, 32, height))


class LevelController:
    def __init__(self, engine):
        self._engine = engine
        self._next_level = None

    def change_level(self, name):
        self._next_level = "data/levels/" + name + ".txt"

    def has_next_level(self):
        return self._next_level is not None

    def commit_change(self):
        if self.has_next_level():
            self._engine.change_level(self._next_level)


class Engine:
    def __init__(self):
        self._reset()

    def _reset(self):
        self.renderables = []
        self.components = []

        self.character = None
        self.mover = None

        self._init_collider()
        self.action = Action()

    def _init_collider(self):
        self.collider = Collider()
        self.collider.register_collidable(LeftBorder())

    def _init_character(self):
        self.character = None
        self.mover = None

        if self.current_level is not None:
            where = self.current_level.search('S')
            if where != (-1, -1):
                character = Character()
                character.teleport(level.position_to_screen(where))
                character.set_collider(self.collider)
                self.renderables.append(get_character_renderer(character, BLUE))
                self.character = character


    def _init_systems(self):
        self.systems = {}
        self.systems['collider'] = self.collider
        self.systems['writer'] = TextWriter()
        self.systems['action'] = self.action
        self.systems['character'] = self.character
        self.systems['level_control'] = LevelController(self)

    def _init_mover(self):
        # Could be extracted as a standard component
        self.mover = Mover(self.character)
        self.mover.on_init(self.systems)
        self.components.append(self.mover)

    def change_level(self, level_path):
        self._reset()

        self.current_level = level.Level().color(WHITE)
        self.current_level.load(level_path)

        self._init_character()
        self._init_systems()
        self._init_mover()

        self.renderables.append(self.current_level)
        for component in self.current_level.get_components():
            component.on_init(self.systems)
            self.components.append(component)

            if callable(getattr(component, "render", None)):
                self.renderables.append(component)


    def update(self, dt):
        self.systems['level_control'].commit_change()

def run():
    engine = Engine()

    engine.change_level("data/levels/000.txt")

    pygame.init()
    clock = pygame.time.Clock()

    window_surface = pygame.display.set_mode(game_surface_size)
    pygame.display.set_caption("Gates")

    key_actions = {}
    key_actions[(pygame.KEYDOWN, pygame.K_RIGHT)] = lambda: engine.mover.start_move((180, 0))
    key_actions[(pygame.KEYUP, pygame.K_RIGHT)] = lambda: engine.mover.stop_move((180, 0))
    key_actions[(pygame.KEYDOWN, pygame.K_LEFT)] = lambda: engine.mover.start_move((-180, 0))
    key_actions[(pygame.KEYUP, pygame.K_LEFT)] = lambda: engine.mover.stop_move((-180, 0))
    key_actions[(pygame.KEYDOWN, pygame.K_SPACE)] = lambda: engine.action.trigger()

    dispatcher = EventDispatcher()
    dispatcher.set_key_actions(key_actions)

    want_to_quit = False
    dt = 0
    while not want_to_quit:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                want_to_quit = True

            dispatcher.process(event)

        for c in engine.components:
            c.update(dt)

        window_surface.fill(BLACK)

        for r in engine.renderables:
            r.render(window_surface)

        pygame.display.flip()
        pygame.display.update()

        engine.update(dt)

        dt = clock.tick(60)

    pygame.display.quit()

if __name__ == '__main__':
    pass

